<?php

namespace MyNamespace;

/**
 * @param Foo $a
 * @param Bar $b
 */
function foo($a, $b)
{

}

class Foo
{

    /**
     * @param Foo $a
     * @param Bar $b
     */
    public function doFoo($a, $b)
    {

    }

}
